import re

import jsonref

from ocdskit.exceptions import MissingColumnError
from ocdskit.schema import get_schema_fields
from ocdskit.util import _cast_as_list

# See https://stackoverflow.com/questions/30734682/extracting-url-and-anchor-text-from-markdown-using-python
INLINE_LINK_RE = re.compile(r'\[([^\]]+)\]\(([^)]+)\)')


def mapping_sheet(schema, order_by=None, infer_required=False, extension_field=None, inherit_extension=True,
                  include_codelist=False, include_deprecated=True, include_definitions=False, base_uri=None):
    """
    Returns information about all field paths in a JSON Schema, as columns and rows.

    If ``include_definitions=False``, this function resolves ``$ref`` properties.

    :param dict schema: a JSON schema
    :param str order_by: the column by which to sort the rows
    :param bool infer_required: whether to infer that a field is required if "null" is not in its ``type``
    :param str extension_field: the property in the JSON schema containing the name of the extension in which each
                                field was defined
    :param bool inherit_extension: whether fields defined via $ref properties inherit the "extension" value of their
                                parent field
    :param bool include_codelist: whether to include a "codelist" column
    :param bool include_deprecated: whether to include any deprecated fields
    :param bool include_definitions: whether to traverse the "definitions" property
    :param str base_uri: the URL to resolve relative references against
    :returns: information about all field paths in a JSON Schema, as columns and rows
    :rtype: tuple

    The columns are:

    :``section``: The first part of the JSON path to the field in the data, e.g. ``tender``
    :``path``: The JSON path to the field in the data, e.g. ``tender/id``
    :``title``: The field's ``title`` in the JSON schema.  If the field has no ``title``, defaults to the field's name
      followed by "*".
    :``description``: The field's ``description`` in the JSON schema. URLs are removed (see the ``links`` column).
    :``type``: A comma-separated list of the field's ``type`` in the JSON schema, excluding "null". If the field has no
      ``type``, defaults to "unknown".
    :``range``: The field's allowed number of occurrences.

      * "0..1" if the field defines an optional literal value.
      * "0..n" if the field defines an optional array.
      * "1..1" if the field defines a required literal value.
      * "1..n" if the field defines a required array.
    :``values``: If the field's schema sets:

      * ``format``: the ``format``
      * ``pattern``: the ``pattern``
      * ``enum``: "Enum: " followed by the ``enum`` as a comma-separated list, excluding ``null``
      * ``items/enum``: "Enum: " followed by the ``items/enum`` as a comma-separated list, excluding ``null``
    :``codelist``: The field's ``codelist`` in the JSON schema
    :``links``: The URLs extracted from the field's ``description``
    :``deprecated``: The OCDS minor version in which the field (or its parent) was deprecated
    :``deprecationNotes``: The explanation for the deprecation of the field
    :``extension``: The name of the extension that introduced the JSON path (see the ``extension_field`` parameter)

    :raises MissingColumnError: if the column by which to order is missing
    """
    kwargs = {
        'inherit_extension': inherit_extension,
        'include_codelist': include_codelist,
        'include_deprecated': include_deprecated,
    }

    if not include_definitions:
        # jsonref.JsonRef is deprecated, but used for backwards-compatibility with jsonref 0.x.
        schema = jsonref.JsonRef.replace_refs(schema, base_uri=base_uri, jsonschema=True)

    rows = []
    rows_by_path = {}
    for field in get_schema_fields(schema):
        if not include_definitions and field.definition_pointer_components:
            continue

        prop = field.schema
        field.sep = '/'
        extension_name = prop.get(extension_field)

        # If the schema sets a `$ref` property, add an extra row for it. This preserves any differences in the titles
        # and descriptions of the referrer and referee. The new row can be formatted as a heading for the object.
        if hasattr(prop, '__reference__'):
            reference = dict(prop.__reference__)
            prop = dict(prop)
            if extension_field in reference:
                extension_name = reference[extension_field]
            if 'type' not in reference and 'type' in prop:
                reference['type'] = prop['type']
            _add_row(rows, rows_by_path, field, reference, extension_name, infer_required=infer_required,
                     **kwargs)

        _add_row(rows, rows_by_path, field, prop, extension_name, infer_required=infer_required,
                 **kwargs)

        # If the field is an array, add an extra row for it. This makes it easier to use as a header for the object.
        if 'items' in prop and 'properties' in prop['items'] and 'title' in prop['items']:
            row = {
                'path': field.path,
                'title': prop['items']['title'],
                'description': prop['items'].get('description', ''),
                'type': prop['items']['type'],
                'deprecated': field.deprecated.get('deprecatedVersion'),  # deprecation from parent
            }
            _add_deprecated(row, prop['items'])

            _add_row(rows, rows_by_path, field, prop['items'], extension_name, row=row, **kwargs)

    if order_by:
        try:
            rows.sort(key=lambda row: row[order_by])
        except KeyError as e:
            raise MissingColumnError(f"the column '{order_by}' doesn't exist – did you make a typo?") from e

    columns = ['section', 'path', 'title', 'description', 'type', 'range', 'values', 'links', 'deprecated',
               'deprecationNotes']
    if extension_field:
        columns.append('extension')
    if include_codelist:
        columns.append('codelist')

    return columns, rows


def _add_deprecated(row, schema):
    # OCDS for PPPs sets `"deprecated": null`.
    if schema.get('deprecated'):
        row['deprecated'] = schema['deprecated'].get('deprecatedVersion', '')
        row['deprecationNotes'] = schema['deprecated'].get('description', '')


def _add_row(rows, rows_by_path, field, schema, extension_name, *, infer_required=None, inherit_extension=True,
             include_codelist=False, include_deprecated=True, row=None):
    parent = rows_by_path.get(field.path_components[:-1], {})
    if not row:
        row = _make_row(field, schema, infer_required, include_codelist)

    if extension_name:
        row['extension'] = extension_name
    elif 'extension' in parent and inherit_extension:
        row['extension'] = parent['extension']

    if include_deprecated or not row['deprecated']:
        rows.append(row)

    rows_by_path[field.path_components] = row


def _make_row(field, schema, infer_required, include_codelist):
    row = {
        'path': field.path,
        'title': schema.get('title', field.path_components[-1] + '*'),
        'deprecated': field.deprecated.get('deprecatedVersion'),  # deprecation from parent
    }

    if len(field.path_components) > 1:
        row['section'] = field.path_components[0]
    else:
        row['section'] = field.definition_path

    if 'description' in schema:
        links = dict(INLINE_LINK_RE.findall(schema['description']))
        row['description'] = schema['description']
        for key, link in links.items():
            row['description'] = row['description'].replace('[' + key + '](' + link + ')', key)
        row['links'] = ', '.join(links.values())

    required = False

    if 'type' in schema:
        types = _cast_as_list(schema['type'])

        if 'null' in types:
            types.remove('null')
        elif infer_required:
            required = 'string' in types or 'integer' in types

        row['type'] = ', '.join(types)
    else:
        row['type'] = 'unknown'

    if field.required:
        required = True

    min_range = '1' if required else '0'
    max_range = 'n' if row['type'] == 'array' else '1'
    row['range'] = f'{min_range}..{max_range}'

    if 'format' in schema:
        row['values'] = schema['format']
    elif 'pattern' in schema:
        row['values'] = 'Pattern: ' + schema['pattern']
    elif 'enum' in schema:
        values = list(schema['enum'])
        if None in values:
            values.remove(None)
        row['values'] = 'Enum: ' + ', '.join(values)
    elif 'items' in schema and 'enum' in schema['items']:
        values = list(schema['items']['enum'])
        if None in values:
            values.remove(None)
        row['values'] = 'Enum: ' + ', '.join(values)
    else:
        row['values'] = ''

    if include_codelist:
        row['codelist'] = schema.get('codelist')

    _add_deprecated(row, schema)

    return row



from ocdskit.commands.base import OCDSCommand


class Command(OCDSCommand):
    name = 'echo'
    help = 'Repeats the input, applying --encoding, --ascii, --pretty and --root-path, and using the UTF-8 encoding'

    def handle(self):
        for data in self.items():
            self.print(data)



from dataclasses import dataclass, field

LANGUAGE_CODE_SUFFIX = '_(((([A-Za-z]{2,3}(-([A-Za-z]{3}(-[A-Za-z]{3}){0,2}))?)|[A-Za-z]{4}|[A-Za-z]{5,8})(-([A-Za-z]{4}))?(-([A-Za-z]{2}|[0-9]{3}))?(-([A-Za-z0-9]{5,8}|[0-9][A-Za-z0-9]{3}))*(-([0-9A-WY-Za-wy-z](-[A-Za-z0-9]{2,8})+))*(-(x(-[A-Za-z0-9]{1,8})+))?)|(x(-[A-Za-z0-9]{1,8})+))'  # noqa: E501
LANGUAGE_CODE_SUFFIX_LEN = len(LANGUAGE_CODE_SUFFIX)


def _join_sep(sep, words):
    return sep + sep.join(words)


@dataclass
class Field:
    """
    Initializes a schema field object.
    """

    #: the field's schema
    schema: dict = None
    #: the JSON pointer to the field in the schema, e.g. ``('properties', 'tender', 'properties', 'id')``
    pointer_components: tuple = None
    #: the path to the field in data, e.g. ``('tender', 'id')``
    path_components: tuple = None
    #: the JSON pointer to the definition in which the field is defined, e.g. ``('definitions', 'Item')``
    definition_pointer_components: tuple = None
    #: the path to the definition in which the field is defined, e.g. ``('Item')``
    definition_path_components: tuple = None
    #: whether the field is listed in the schema's ``required``
    required: bool = None
    #: if the field, or an ancestor of the field, sets ``deprecated``, the ``deprecated`` object
    deprecated: dict = field(default_factory=dict)
    #: whether the field has a corresponding field in the schema's ``patternProperties``
    multilingual: bool = None
    #: the separator to use in string representations of paths
    sep = '.'

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    @property
    def pointer(self):
        """
        Returns the JSON pointer to the field in the schema, e.g. ``/properties/tender/properties/id``.
        """
        return _join_sep('/', self.pointer_components)

    @property
    def definition_pointer(self):
        """
        Returns the JSON pointer to the definition in which the field is defined, e.g. ``/definitions/Item``.
        """
        return _join_sep('/', self.definition_pointer_components)

    @property
    def path(self):
        """
        Returns the path to the field in data with ``self.sep`` as separator, e.g. ``tender.id``.
        """
        return self.sep.join(self.path_components)

    @property
    def definition_path(self):
        """
        Returns the path to the definition in which the field is defined with ``self.sep`` as separator, e.g. ``Item``.
        """
        return self.sep.join(self.definition_path_components)

    def __repr__(self):
        return repr(self.asdict())

    def asdict(self, sep=None, exclude=None):
        """
        Returns the field as a dict, with keys for: ``schema``, ``pointer``, ``path``,
        ``definition_pointer``, ``definition_path``, ``required``, ``deprecated``, ``multilingual``.

        :param list sep: the separator to use in string representations of paths, overriding ``self.sep``
        :param list exclude: a list of keys to exclude from the dict
        """
        data = {}

        exclude = exclude or ()
        sep = sep or self.sep

        for key, value in self.__dict__.items():
            if key not in exclude and not key.startswith('_') and not key.endswith('_components'):
                data[key] = value
        for key in ('pointer', 'definition_pointer'):
            if key not in exclude:
                data[key] = getattr(self, key)
        for key in ('path', 'definition_path'):
            if key not in exclude:
                data[key] = sep.join(getattr(self, f'{key}_components'))

        return data


# This code is similar to `add_versioned` in `make_versioned_release_schema.py` in the `standard` repository.
def get_schema_fields(schema, pointer=None, path=None, definition_pointer=None, definition_path=None, deprecated=None):
    """
    Yields a ``Field`` object for each field (whether under ``properties`` or ``patternProperties``) in a JSON schema.

    :param dict schema: a JSON schema
    :param tuple pointer: the JSON pointer to the field in the schema, e.g.
                          ``('properties', 'tender', 'properties', 'id')``
    :param tuple path: the path to the field in data, e.g.
                       ``('tender', 'id')``
    :param tuple definition_pointer: the JSON pointer to the definition in which the field is defined, e.g.
                                     ``('definitions', 'Item')``
    :param tuple definition_path: the path to the definition in which the field is defined, e.g.
                                  ``('Item')``
    :param dict deprecated: if the field, or an ancestor of the field, sets ``deprecated``, the ``deprecated``
                            object
    """
    if pointer is None:
        pointer = ()
    if path is None:
        path = ()
    if definition_pointer is None:
        definition_pointer = ()
    if definition_path is None:
        definition_path = ()

    multilingual = set()
    hidden = set()
    for key, value in schema.get('patternProperties', {}).items():
        # The pattern might have an extra set of parentheses.
        for offset in (2, 1):
            end = -LANGUAGE_CODE_SUFFIX_LEN - offset
            # The pattern must be anchored and the suffix must occur at the end.
            if (
                key[end:-offset] == LANGUAGE_CODE_SUFFIX
                and key[:offset] == '^('[:offset]
                and key[-offset:] == ')$'[-offset:]
            ):
                multilingual.add(key[offset:end])
                hidden.add(key)
                break

    for key, value in schema.get('properties', {}).items():
        new_pointer = pointer + ('properties', key)
        new_path = path + (key,)
        required = schema.get('required', [])
        yield from _get_schema_field(key, value, new_pointer, new_path, definition_pointer, definition_path,
                                     required, deprecated or _deprecated(value), multilingual)

    for key, value in schema.get('definitions', {}).items():
        new_pointer = pointer + ('definitions', key)
        yield from get_schema_fields(value, pointer=new_pointer, path=(), definition_pointer=new_pointer,
                                     definition_path=(key,), deprecated=deprecated)

    for key, value in schema.get('patternProperties', {}).items():
        if key not in hidden:
            new_pointer = pointer + ('patternProperties', key)
            new_path = path + (f'({key})',)
            yield Field(schema=value, pointer_components=new_pointer, path_components=new_path,
                        definition_pointer_components=definition_pointer, definition_path_components=definition_path,
                        required=False, deprecated=deprecated or _deprecated(value), multilingual=False)


def _get_schema_field(name, schema, pointer, path, definition_pointer, definition_path, required, deprecated,
                      multilingual):
    yield Field(schema=schema, pointer_components=pointer, path_components=path,
                definition_pointer_components=definition_pointer, definition_path_components=definition_path,
                required=name in required, deprecated=deprecated, multilingual=name in multilingual)

    if schema is None:
        return

    if 'properties' in schema or 'patternProperties' in schema:
        yield from get_schema_fields(schema, pointer=pointer, path=path, definition_pointer=definition_pointer,
                                     definition_path=definition_path, deprecated=deprecated)

    for key, value in schema.get('items', {}).get('properties', {}).items():
        new_pointer = pointer + ('items', 'properties', key)
        new_path = path + (key,)
        required = schema['items'].get('required', [])
        yield from _get_schema_field(key, value, new_pointer, new_path, definition_pointer, definition_path,
                                     required, deprecated or _deprecated(value), multilingual)


def _deprecated(value):
    if value is None:
        return {}
    return value.get('deprecated') or hasattr(value, '__reference__') and value.__reference__.get('deprecated') or {}


def add_validation_properties(schema, unique_items=True, coordinates=False):
    """
    Adds "minItems" and "uniqueItems" if an array, adds "minProperties" if an object, and adds "minLength" if a string
    and if "enum", "format" and "pattern" aren't set.

    :param dict schema: a JSON schema
    :param bool unique_items: whether to add "uniqueItems" properties to array fields
    :param bool coordinates: whether the parent is a geospatial coordinates field
    """
    if isinstance(schema, list):
        for item in schema:
            add_validation_properties(item, unique_items=unique_items)
    elif isinstance(schema, dict):
        if 'type' in schema:
            if (
                'string' in schema['type']
                # "enum" is more strict than "minLength".
                and 'enum' not in schema
                # The defined formats do not match zero-length strings.
                # https://datatracker.ietf.org/doc/html/draft-fge-json-schema-validation-00#section-7.3
                and 'format' not in schema
                # The pattern is assumed to not match zero-length strings.
                and 'pattern' not in schema
            ):
                schema.setdefault('minLength', 1)

            if 'array' in schema['type']:
                # Allow non-unique items for coordinates fields (e.g. closed polygons).
                if sorted(schema.get('items', {}).get('type', [])) == ['array', 'number']:
                    coordinates = True
                if unique_items and not coordinates:
                    schema.setdefault('uniqueItems', True)
                schema.setdefault('minItems', 1)

            if 'object' in schema['type']:
                schema.setdefault('minProperties', 1)

        for value in schema.values():
            add_validation_properties(value, unique_items=unique_items, coordinates=coordinates)



from ocdskit.__main__ import main
from tests import assert_streaming


def test_command(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['split-record-packages', '1'],
                     ['realdata/record-package_package.json'], ['realdata/record-package_split.json'])



import json
import logging

import pytest

import ocdskit.combine
from ocdskit.__main__ import main
from ocdskit.util import json_dumps
from tests import assert_streaming, assert_streaming_error, read, run_streaming


def _remove_package_metadata(filenames):
    outputs = []
    for filename in filenames:
        data = json.loads(read(filename))
        data['publisher'] = {}
        del data['packages']
        outputs.append(data)
    return json_dumps(data) + '\n'


# Test with packages and with releases.
def assert_compile_command(capsys, monkeypatch, main, args, stdin, expected, remove_package_metadata=False):
    assert_streaming(capsys, monkeypatch, main, args, stdin, expected)

    args[args.index('compile') + 1:0] = ['--root-path', 'releases.item']
    if remove_package_metadata:
        expected = _remove_package_metadata(expected)
    assert_streaming(capsys, monkeypatch, main, args, stdin, expected)


@pytest.mark.usefixtures('sqlite')
def test_command(capsys, monkeypatch):
    assert_compile_command(capsys, monkeypatch, main, ['--ascii', 'compile'],
                           ['realdata/release-package-1.json', 'realdata/release-package-2.json'],
                           ['realdata/compiled-release-1.json', 'realdata/compiled-release-2.json'])


@pytest.mark.usefixtures('sqlite')
def test_command_extensions_with_packages(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['compile'],
                     ['release-package_additional-contact-points.json'], ['compile_extensions.json'])


@pytest.mark.usefixtures('sqlite')
def test_command_extensions_with_releases(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['compile', '--root-path', 'releases.item'],
                     ['release-package_additional-contact-points.json'], ['compile_no-extensions.json'])


@pytest.mark.usefixtures('sqlite')
def test_command_versioned(capsys, monkeypatch):
    assert_compile_command(capsys, monkeypatch, main, ['--ascii', 'compile', '--versioned'],
                           ['realdata/release-package-1.json', 'realdata/release-package-2.json'],
                           ['realdata/versioned-release-1.json', 'realdata/versioned-release-2.json'])


@pytest.mark.usefixtures('sqlite')
def test_command_package(capsys, monkeypatch):
    assert_compile_command(capsys, monkeypatch, main, ['compile', '--package'],
                           ['realdata/release-package-1.json', 'realdata/release-package-2.json'],
                           ['realdata/record-package_package.json'], remove_package_metadata=True)


@pytest.mark.usefixtures('sqlite')
def test_command_package_uri_published_date_version(capsys, monkeypatch):
    actual = run_streaming(capsys, monkeypatch, main, ['compile', '--package', '--uri', 'http://example.com/x.json',
                                                       '--published-date', '2010-01-01T00:00:00Z', '--version', 'X'],
                           ['release-package_minimal.json'])

    package = json.loads(actual.out)
    assert package['uri'] == 'http://example.com/x.json'
    assert package['publishedDate'] == '2010-01-01T00:00:00Z'
    assert package['version'] == 'X'


@pytest.mark.usefixtures('sqlite')
def test_command_package_publisher(capsys, monkeypatch):
    actual = run_streaming(capsys, monkeypatch, main, ['compile', '--package', '--publisher-name', 'Acme Inc.',
                                                       '--publisher-uri', 'http://example.com/', '--publisher-scheme',
                                                       'scheme', '--publisher-uid', '12345'],
                           ['release-package_minimal.json'])

    package = json.loads(actual.out)
    assert package['publisher']['name'] == 'Acme Inc.'
    assert package['publisher']['uri'] == 'http://example.com/'
    assert package['publisher']['scheme'] == 'scheme'
    assert package['publisher']['uid'] == '12345'


@pytest.mark.usefixtures('sqlite')
def test_command_package_fake(capsys, monkeypatch):
    actual = run_streaming(capsys, monkeypatch, main, ['compile', '--package', '--fake'],
                           ['release-package_minimal.json'])

    package = json.loads(actual.out)
    assert package['uri'] == 'placeholder:'
    assert package['publishedDate'] == '9999-01-01T00:00:00Z'


@pytest.mark.usefixtures('sqlite')
def test_command_package_packages(capsys, monkeypatch):
    actual = run_streaming(capsys, monkeypatch, main, ['compile', '--package'],
                           ['release_minimal.json'])

    package = json.loads(actual.out)
    assert 'packages' not in package


@pytest.mark.usefixtures('sqlite')
def test_command_warning(capsys, monkeypatch, caplog):
    actual = run_streaming(capsys, monkeypatch, main, ['compile'],
                           ['release-package_warning.json'])

    assert actual.out == read('compile_warning.json')
    assert actual.err == "ocds-213czf-1: Multiple objects have the `id` value '1' in the `parties` array\n"


@pytest.mark.usefixtures('sqlite')
def test_command_package_linked_releases_with_packages(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['compile', '--package', '--linked-releases'],
                     ['realdata/release-package-1.json', 'realdata/release-package-2.json'],
                     ['realdata/record-package_linked-releases.json'])


@pytest.mark.usefixtures('sqlite')
def test_command_package_linked_releases_with_releases(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['compile', '--package', '--linked-releases', '--root-path',
                                                 'releases.item'],
                     ['realdata/release-package-1.json', 'realdata/release-package-2.json'],
                     _remove_package_metadata(['realdata/record-package_package.json']))


@pytest.mark.usefixtures('sqlite')
def test_command_package_versioned(capsys, monkeypatch):
    assert_compile_command(capsys, monkeypatch, main, ['compile', '--package', '--versioned'],
                           ['realdata/release-package-1.json', 'realdata/release-package-2.json'],
                           ['realdata/record-package_versioned.json'], remove_package_metadata=True)


@pytest.mark.usefixtures('sqlite')
def test_command_version_mismatch(capsys, monkeypatch, caplog):
    with caplog.at_level(logging.ERROR):
        assert_streaming_error(capsys, monkeypatch, main, ['compile', '--package', '--versioned'],
                               ['realdata/release-package_1.1-1.json', 'realdata/release-package_1.0-1.json'])

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'CRITICAL'
        assert caplog.records[0].message == "item 1: version error: this item uses version 1.0, but earlier items " \
            "used version 1.1\nTry first upgrading items to the same version:\n  cat file [file ...] | ocdskit " \
            "upgrade 1.0:1.1 | ocdskit compile --package --versioned"


@pytest.mark.usefixtures('sqlite')
def test_command_missing_ocid(capsys, monkeypatch, caplog):
    stdin = b'{"id":"1","date":"2001-02-03T04:05:06Z","tag":["planning"],"initiationType":"tender"}'

    with caplog.at_level(logging.ERROR):
        assert_streaming_error(capsys, monkeypatch, main, ['compile'], stdin)

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'CRITICAL'
        assert caplog.records[0].message == 'The `ocid` field of at least one release is missing.'


@pytest.mark.usefixtures('sqlite')
def test_command_unknown_version(capsys, monkeypatch, caplog):
    with caplog.at_level(logging.ERROR):
        assert_streaming_error(capsys, monkeypatch, main, ['compile'], ['release-package_unknown-version.json'])

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'CRITICAL'
        assert caplog.records[0].message == 'The `version` value ("X") of a release package is not recognized.'


def test_command_without_sqlite(capsys, monkeypatch, caplog):
    ocdskit.combine.sqlite = False

    # To check the warning, not the output.
    run_streaming(capsys, monkeypatch, main, ['compile'], ['release-package_minimal.json'])

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[0].message == 'sqlite3 is unavailable, so the command will run in memory. If input files ' \
                                        'are too large, the command might exceed available memory.'



import json
import logging
from collections import OrderedDict  # for move_to_end()
from copy import deepcopy
from hashlib import md5

from ocdskit.util import (
    _cast_as_list,
    get_ocds_minor_version,
    is_package,
    is_record,
    is_record_package,
    is_release_package,
)

logger = logging.getLogger('ocdskit')

# See https://standard.open-contracting.org/1.0/en/schema/reference/#identifier
organization_identification_1_0 = (
    (None, ('name',)),
    ('identifier', ('scheme', 'id', 'legalName', 'uri')),
    ('address', ('streetAddress', 'locality', 'region', 'postalCode', 'countryName')),
    ('contactPoint', ('name', 'email', 'telephone', 'faxNumber', 'url')),
)


def _move_to_top(data, fields):
    for field in reversed(fields):
        if field in data:
            data.move_to_end(field, last=False)


def _in(obj, field):
    return field in obj and obj[field] is not None


def upgrade_10_10(data):
    """
    Upgrades a record package, release package, record or release from 1.0 to 1.0 (no-op).
    """
    return data


def upgrade_11_11(data):
    """
    Upgrades a record package, release package, record or release from 1.1 to 1.1 (no-op).
    """
    return data


def upgrade_10_11(data):
    """
    Upgrades a record package, release package, record or release from 1.0 to 1.1.

    Retains the deprecated Amendment.changes, Budget.source and Milestone.documents fields.

    ``data`` must be an ``OrderedDict``. If you have only the parsed JSON, re-parse it with:

    ``upgrade_10_11(json.loads(json.dumps(data), object_pairs_hook=OrderedDict))``
    """
    version = get_ocds_minor_version(data)
    if version != '1.0':
        return data

    if is_package(data):
        data['version'] = '1.1'
        _move_to_top(data, ('uri', 'version'))

    if is_record_package(data):
        for record in data['records']:
            upgrade_record_10_11(record)
    elif is_release_package(data):
        for release in data['releases']:
            upgrade_release_10_11(release)
    elif is_record(data):
        upgrade_record_10_11(data)
    else:  # release
        upgrade_release_10_11(data)

    return data


def upgrade_record_10_11(record):
    """
    Upgrades a record from 1.0 to 1.1.
    """
    if 'releases' in record:
        for release in record['releases']:
            upgrade_release_10_11(release)
    if 'compiledRelease' in record:
        upgrade_release_10_11(record['compiledRelease'])


def upgrade_release_10_11(release):
    """
    Applies upgrades for organization handling, amendment handling and transactions terminology.
    """
    upgrade_parties_10_to_11(release)
    upgrade_amendments_10_11(release)
    upgrade_transactions_10_11(release)


def upgrade_parties_10_to_11(release):
    """
    Converts organizations to organization references and fills in the ``parties`` array.
    """
    parties = _get_parties(release)

    if _in(release, 'buyer'):
        buyer = release['buyer']
        release['buyer'] = _add_party(parties, buyer, 'buyer')

    if _in(release, 'tender'):
        if _in(release['tender'], 'procuringEntity'):
            procuring_entity = release['tender']['procuringEntity']
            release['tender']['procuringEntity'] = _add_party(parties, procuring_entity, 'procuringEntity')
        if _in(release['tender'], 'tenderers'):
            for i, tenderer in enumerate(release['tender']['tenderers']):
                release['tender']['tenderers'][i] = _add_party(parties, tenderer, 'tenderer')

    if _in(release, 'awards'):
        for award in release['awards']:
            if _in(award, 'suppliers'):
                for i, supplier in enumerate(award['suppliers']):
                    award['suppliers'][i] = _add_party(parties, supplier, 'supplier')

    if parties:
        if 'parties' not in release:
            release['parties'] = []
            _move_to_top(release, ('ocid', 'id', 'date', 'tag', 'initiationType', 'parties'))

        for party in parties.values():
            if party not in release['parties']:
                release['parties'].append(party)


def _get_parties(release):
    parties = {}

    if 'parties' in release:
        for party in release['parties']:
            if 'id' in party:
                _id = party['id']
            else:
                _id = _create_party_id(party)
            parties[_id] = party

    return parties


def _add_party(parties, party, role):
    """
    Adds an ``id`` to the party, adds the party to the ``parties`` array, sets the party's role, and returns an
    OrganizationReference. Warns if there is any data loss from differences in non-identifying fields.
    """
    party = deepcopy(party)

    if 'id' not in party:
        party['id'] = _create_party_id(party)
        _move_to_top(party, ('id'))

    _id = party['id']

    if _id not in parties:
        parties[_id] = party
    else:
        # Warn about data loss.
        survivor = deepcopy(parties[_id])
        roles = survivor.pop('roles', [])
        if any(key not in survivor or party[key] != survivor[key] for key in party):
            logger.warning('party in "%s" role differs from party in %s roles:\n%s\n%s', role, json.dumps(roles),
                           json.dumps(party), json.dumps(survivor))

    if 'roles' not in parties[_id]:
        parties[_id]['roles'] = []
        _move_to_top(parties[_id], ('id', 'roles'))

    # In case the data is invalid.
    parties[_id]['roles'] = _cast_as_list(parties[_id]['roles'])

    if role not in parties[_id]['roles']:
        # Update the `roles` of the party in the `parties` array.
        parties[_id]['roles'].append(role)

    # Create the OrganizationReference.
    organization_reference = {'id': _id}

    if 'name' in party:
        organization_reference['name'] = party['name']

    return organization_reference


def _create_party_id(party):
    parts = []
    for parent, fields in organization_identification_1_0:
        if not parent:
            for field in fields:
                parts.append(_get_bytes(party, field))
        elif parent in party:
            for field in fields:
                parts.append(_get_bytes(party[parent], field))

    return md5(b'-'.join(parts)).hexdigest()


def _get_bytes(obj, field):
    # Handle null and integers.
    return str(obj.get(field) or '').encode('utf-8')


def upgrade_amendments_10_11(release):
    """
    Renames ``amendment`` to ``amendments`` under ``tender``, ``awards`` and ``contracts``. If ``amendments`` already
    exists, it appends the ``amendment`` value to the ``amendments`` array, unless it already contains it.
    """
    if _in(release, 'tender'):
        _upgrade_amendment_10_11(release['tender'])
    for field in ('awards', 'contracts'):
        if _in(release, field):
            for block in release[field]:
                _upgrade_amendment_10_11(block)


def _upgrade_amendment_10_11(block):
    if 'amendment' in block:
        block.setdefault('amendments', [])
        if block['amendment'] not in block['amendments']:
            block['amendments'].append(block['amendment'])
        del block['amendment']


def upgrade_transactions_10_11(release):
    """
    Renames ``providerOrganization`` to ``payer``, ``receiverOrganization`` to ``payee``, and ``amount`` to ``value``
    under ``contracts.implementation.transactions``, unless they already exist.

    Converts ``providerOrganization`` and ``receiverOrganization`` from an Identifier to an OrganizationReference and
    fills in the ``parties`` array.
    """
    parties = _get_parties(release)

    if _in(release, 'contracts'):
        for contract in release['contracts']:
            if _in(contract, 'implementation') and _in(contract['implementation'], 'transactions'):
                for transaction in contract['implementation']['transactions']:
                    if 'value' not in transaction:
                        transaction['value'] = transaction['amount']
                        del transaction['amount']

                    for old, new in (('providerOrganization', 'payer'), ('receiverOrganization', 'payee')):
                        if old in transaction and new not in transaction:
                            party = OrderedDict([('identifier', transaction[old])])

                            if 'legalName' in transaction[old]:
                                party['name'] = transaction[old]['legalName']

                            transaction[new] = _add_party(parties, party, new)
                            del transaction[old]



from ocdskit.__main__ import main
from tests import assert_streaming


def test_command(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['split-release-packages', '2'], ['realdata/release-package-1-2.json'],
                     ['realdata/release-package_split.json'])



from ocdskit.__main__ import main
from tests import path, run_command


def test_command(capsys, monkeypatch):
    actual = run_command(capsys, monkeypatch, main, ['schema-report', '--min-occurrences', '2',
                                                     path('test-schema.json')])

    assert actual.out == 'codelist,openCodelist\n' \
                         'a.csv,False/True\n' \
                         'b.csv,False\n' \
                         'c.csv,False\n' \
                         'd.csv,False\n' \
                         '\n' \
                         " 2: {'codelist': 'a.csv', 'openCodelist': True, 'type': ['string', 'null']}\n"


def test_command_no_codelists(capsys, monkeypatch):
    actual = run_command(capsys, monkeypatch, main, ['schema-report', '--min-occurrences', '2', '--no-codelists',
                                                     path('test-schema.json')])

    assert 'codelist,openCodelist' not in actual.out
    assert ':' in actual.out


def test_command_no_definitions(capsys, monkeypatch):
    actual = run_command(capsys, monkeypatch, main, ['schema-report', '--min-occurrences', '2', '--no-definitions',
                                                     path('test-schema.json')])

    assert 'codelist,openCodelist' in actual.out
    assert ':' not in actual.out


def test_command_min_occurrences(capsys, monkeypatch):
    actual = run_command(capsys, monkeypatch, main, ['schema-report', '--min-occurrences', '1', '--no-codelists',
                                                     path('test-schema.json')])

    assert 'codelist,openCodelist' not in actual.out
    assert '1:' in actual.out



import logging
import sys

import ocdskit.packager
from ocdskit.combine import merge
from ocdskit.commands.base import OCDSCommand
from ocdskit.exceptions import CommandError, InconsistentVersionError, MissingOcidKeyError, UnknownVersionError

logger = logging.getLogger('ocdskit')


class Command(OCDSCommand):
    name = 'compile'
    help = 'reads release packages and individual releases from standard input, merges the releases by OCID, and ' \
           'prints the compiled releases'

    def add_arguments(self):
        self.add_argument('--schema', help='the URL or path of the patched release schema to use')
        self.add_argument('--package', action='store_true', help='wrap the compiled releases in a record package')
        self.add_argument('--linked-releases', action='store_true',
                          help='if --package is set, use linked releases instead of full releases')
        self.add_argument('--versioned', action='store_true',
                          help='if --package is set, include versioned releases in the record package; otherwise, '
                               'print versioned releases instead of compiled releases')

        self.add_package_arguments('record', 'if --package is set, ')

    def handle(self):
        kwargs = self.parse_package_arguments()
        kwargs['schema'] = self.args.schema
        kwargs['return_package'] = self.args.package
        kwargs['use_linked_releases'] = self.args.linked_releases
        kwargs['return_versioned_release'] = self.args.versioned

        if not ocdskit.packager.USING_SQLITE:
            logger.warning('sqlite3 is unavailable, so the command will run in memory. If input files are too large, '
                           'the command might exceed available memory.')

        try:
            for output in merge(self.items(), streaming=True, **kwargs):
                self.print(output, streaming=self.args.package)
        except MissingOcidKeyError as e:
            raise CommandError('The `ocid` field of at least one release is missing.') from e
        except UnknownVersionError as e:
            raise CommandError(f'The `version` value ("{e}") of a release package is not recognized.') from e
        except InconsistentVersionError as e:
            versions = [e.earlier_version, e.current_version]
            if versions[1] < versions[0]:
                versions.reverse()

            raise CommandError(
                f"{e}\nTry first upgrading items to the same version:\n  cat file [file ...] | ocdskit upgrade "
                f"{versions[0]}:{versions[1]} | ocdskit {' '.join(sys.argv[1:])}"
            ) from e



import json

import pytest
from ocdsextensionregistry import ProfileBuilder
from ocdsmerge.exceptions import DuplicateIdValueWarning

from ocdskit.combine import compile_release_packages, merge, package_records
from tests import read


def test_package_default_arguments():
    data = [item for filename in ('realdata/record-package-1.json', 'realdata/record-package-2.json')
            for item in json.loads(read(filename))['records']]

    actual = package_records(data)

    assert actual == json.loads(read('realdata/record-package_record-package.json'))


def test_merge_empty():
    compiled_releases = list(merge([]))

    assert compiled_releases == []


def test_merge_with_schema():
    builder = ProfileBuilder('1__1__4', {'additionalContactPoint': 'master'})
    schema = builder.patched_release_schema()

    data = json.loads(read('release-package_additional-contact-points.json'))['releases']
    compiled_release = list(merge(data, schema=schema))[0]

    assert compiled_release == json.loads(read('compile_extensions.json'))


def test_merge_without_schema():
    data = json.loads(read('release-package_additional-contact-points.json'))['releases']
    compiled_release = list(merge(data))[0]

    assert compiled_release == json.loads(read('compile_no-extensions.json'))


def test_merge_warning():
    data = json.loads(read('release-package_warning.json'))['releases']

    with pytest.warns(DuplicateIdValueWarning) as records:
        compiled_release = list(merge(data))[0]

    assert compiled_release == json.loads(read('compile_warning.json'))

    assert [record.message for record in records] == [
        ("ocds-213czf-1: Multiple objects have the `id` value '1' in the `parties` array"),
    ]


def test_compile_release_packages():
    with pytest.warns(DeprecationWarning) as records:
        compiled_releases = list(compile_release_packages([]))

    assert compiled_releases == []
    assert len(records) == 1
    assert str(records[0].message) == 'compile_release_packages() is deprecated. Use merge() instead.'



from ocdskit.combine import combine_release_packages
from ocdskit.commands.base import OCDSCommand


class Command(OCDSCommand):
    name = 'combine-release-packages'
    help = 'reads release packages from standard input, collects releases, and prints one release package'

    def add_arguments(self):
        self.add_package_arguments('release')

    def handle(self):
        kwargs = self.parse_package_arguments()

        output = combine_release_packages(self.items(), **kwargs)

        self.print(output)



import csv
import json
import pathlib
import sys
from argparse import RawDescriptionHelpFormatter
from textwrap import dedent

from ocdsextensionregistry import ProfileBuilder

from ocdskit.commands.base import BaseCommand
from ocdskit.exceptions import CommandError, MissingColumnError
from ocdskit.mapping_sheet import mapping_sheet


class Command(BaseCommand):
    name = 'mapping-sheet'
    help = 'generates a spreadsheet with all field paths in a JSON Schema'
    kwargs = {
        'epilog': dedent(
            """
            The --extension option must be declared after the file argument. It accepts multiple values, which can be
            extensions' metadata URLs, base URLs and/or download URLs. For example:

                ocdskit mapping-sheet release-schema.json --extension \\
                    https://raw.githubusercontent.com/open-contracting-extensions/ocds_coveredBy_extension/master/extension.json \\
                    https://raw.githubusercontent.com/open-contracting-extensions/ocds_options_extension/master/ \\
                    https://github.com/open-contracting-extensions/ocds_techniques_extension/archive/master.zip \\
                    > mapping-sheet.csv

            The --extension-field option can be used with or without the --extension option.

            - If the --extension option is set, then the --extension-field option may be set to any value. In all
              cases, the result is a mapping sheet with an "extension" column, containing the name of the extension in
              which each field was defined.

            - If the --extension option is not set, then the --extension-field option must be set to the property in
              the JSON schema containing the name of the extension in which each field was defined. If there is no such
              property, then the result is a mapping sheet with no values in its "extension" column.
            """  # noqa: E501
        ),
        'formatter_class': RawDescriptionHelpFormatter,
    }

    def add_arguments(self):
        self.add_argument('file', help='the schema file')
        self.add_argument('--order-by', help="sort the spreadsheet's rows by this column")
        self.add_argument('--infer-required', action='store_true', help='infer whether fields are required')
        self.add_argument('--extension', nargs='*', help='patch the release schema with this extension')
        self.add_argument('--extension-field', help='add an "extension" column for the name of the extension in which '
                          'each field was defined')
        self.add_argument('--no-inherit-extension', action='store_true', help='fields defined via $ref properties do '
                          'not inherit the "extension" value of their parent field')
        self.add_argument('--language', default='en', help='the language to use for the name of the extension')
        self.add_argument('--codelist', action='store_true', help='add a "codelist" column')
        self.add_argument('--no-deprecated', action='store_true', help="don't include deprecated fields")
        self.add_argument('--no-replace-refs', action='store_true', help="don't replace schema with $ref properties "
                          'with the referenced schema')

    def handle(self):
        with open(self.args.file) as f:
            schema = json.load(f)

        if self.args.extension:
            builder = ProfileBuilder(None, self.args.extension)
            schema = builder.patched_release_schema(
                schema=schema, extension_field=self.args.extension_field, language=self.args.language
            )

        try:
            fieldnames, rows = mapping_sheet(
                schema,
                order_by=self.args.order_by,
                infer_required=self.args.infer_required,
                extension_field=self.args.extension_field,
                inherit_extension=not self.args.no_inherit_extension,
                include_codelist=self.args.codelist,
                include_deprecated=not self.args.no_deprecated,
                include_definitions=self.args.no_replace_refs,
                base_uri=pathlib.Path(self.args.file).resolve().as_uri(),
            )

            writer = csv.DictWriter(sys.stdout, fieldnames)
            writer.writeheader()
            writer.writerows(rows)
        except MissingColumnError as e:
            raise CommandError(str(e)) from e



from collections import OrderedDict  # for move_to_end()

from ocdskit import upgrade
from ocdskit.commands.base import OCDSCommand
from ocdskit.exceptions import CommandError


class Command(OCDSCommand):
    name = 'upgrade'
    help = 'upgrades packages, records and releases from an old version of OCDS to a new version'

    def add_arguments(self):
        self.add_argument('versions', help='the colon-separated old and new versions')

    def handle(self):
        versions = self.args.versions

        version_from, version_to = versions.split(':')
        if version_from < version_to:
            direction = 'up'
        else:
            direction = 'down'

        try:
            upgrade_method = getattr(upgrade, f"upgrade_{versions.replace('.', '').replace(':', '_')}")
        except AttributeError as e:
            message = f"{direction}grade from {versions.replace(':', ' to ')} is not supported"
            raise CommandError(message) from e

        for data in self.items(map_type=OrderedDict):
            data = upgrade_method(data)
            self.print(data)



import platform

import pytest

from ocdskit.__main__ import main
from tests import assert_command, assert_command_error, path


def test_command(capsys, monkeypatch):
    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', path('release-schema.json')],
                   'mapping-sheet.csv')


def test_command_codelist(capsys, monkeypatch):
    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--codelist', path('release-schema.json')],
                   'mapping-sheet_codelist.csv')


def test_command_no_deprecated(capsys, monkeypatch):
    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--no-deprecated', path('release-schema.json')],
                   'mapping-sheet_no-deprecated.csv')


def test_command_order_by(capsys, monkeypatch):
    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--order-by', 'path', path('release-schema.json')],
                   'mapping-sheet_order-by.csv')


def test_command_extension(capsys, monkeypatch):
    url = 'https://github.com/open-contracting-extensions/ocds_lots_extension/archive/v1.1.4.zip'

    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', path('release-schema.json'), '--extension', url],
                   'mapping-sheet_extension.csv')


def test_command_extension_field(capsys, monkeypatch):
    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--extension-field', 'extension',
                    path('release-schema.json')],
                   'mapping-sheet_extension-field.csv')


def test_command_extension_and_extension_field(capsys, monkeypatch):
    url = 'https://github.com/open-contracting-extensions/ocds_lots_extension/archive/v1.1.4.zip'

    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--extension-field', 'extension',
                    path('release-schema.json'), '--extension', url],
                   'mapping-sheet_extension_extension-field.csv')


def test_command_extension_and_extension_field_alternative(capsys, monkeypatch):
    url = 'https://github.com/open-contracting-extensions/ocds_lots_extension/archive/v1.1.4.zip'

    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--extension-field', 'alternative',
                    path('release-schema.json'), '--extension', url],
                   'mapping-sheet_extension_extension-field.csv')


def test_command_extension_and_extension_field_and_no_inherit_extension(capsys, monkeypatch):
    url = 'https://github.com/open-contracting-extensions/ocds_lots_extension/archive/v1.1.4.zip'

    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--extension-field', 'extension',
                    path('release-schema.json'), '--extension', url, '--no-inherit-extension'],
                   'mapping-sheet_extension_extension-field_no-inherit-extension.csv')


@pytest.mark.skipif(
    platform.system() in ('Darwin', 'Linux') and platform.python_implementation() == 'PyPy',
    reason='CI outputs /opt/hostedtoolcache/PyPy/3.9.18/x64/lib/pypy3.9/site-packages/certifi/cacert.pem None'
)
def test_command_extension_and_extension_field_and_language(capsys, monkeypatch):
    url = 'https://extensions.open-contracting.org/es/extensions/lots/v1.1.5/'

    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--extension-field', 'extension',
                    path('release-schema.json'), '--extension', url, '--language', 'es'],
                   'mapping-sheet_extension_extension-field_language.csv')


def test_command_extension_and_extension_field_location(capsys, monkeypatch):
    url = 'https://github.com/open-contracting-extensions/ocds_location_extension/archive/v1.1.4.zip'

    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--extension-field', 'extension',
                    path('release-schema.json'), '--extension', url],
                   'mapping-sheet_extension_extension-field_location.csv')


def test_command_extension_and_extension_field_array(capsys, monkeypatch):
    url = 'https://github.com/open-contracting-extensions/ocds_additionalContactPoints_extension/archive/master.zip'

    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--infer-required', '--extension-field', 'extension',
                    path('release-schema.json'), '--extension', url],
                   'mapping-sheet_extension_extension-field_array.csv')


def test_command_oc4ids(capsys, monkeypatch):
    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', path('project-schema.json')],
                   'mapping-sheet_oc4ids.csv')


def test_command_bods(capsys, monkeypatch):
    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', '--order-by', 'path', path('bods/person-statement.json')],
                   'mapping-sheet_bods.csv')


def test_command_sedl(capsys, monkeypatch):
    assert_command(capsys, monkeypatch, main,
                   ['mapping-sheet', path('sedl-schema.json')],
                   'mapping-sheet_sedl.csv')


def test_command_order_by_nonexistent(capsys, monkeypatch, caplog):
    assert_command_error(capsys, monkeypatch, main, ['mapping-sheet', '--order-by',
                                                     'nonexistent', path('release-schema.json')])

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'CRITICAL'
    assert caplog.records[0].message == "the column 'nonexistent' doesn't exist – did you make a typo?"



import logging
import re
import sys
from io import BytesIO, TextIOWrapper
from unittest.mock import patch

import pytest

from ocdskit.__main__ import main
from tests import assert_streaming, assert_streaming_error, read, run_streaming


def test_help(capsys, monkeypatch, caplog):
    stdin = read('release-package_minimal.json', 'rb')

    with pytest.raises(SystemExit) as excinfo:
        with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))):
            monkeypatch.setattr(sys, 'argv', ['ocdskit', '--help'])
            main()

    assert capsys.readouterr().out.startswith('usage: ocdskit [-h] ')

    assert len(caplog.records) == 0
    assert excinfo.value.code == 0


def test_command_encoding(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['--encoding', 'iso-8859-1', 'echo'],
                     ['realdata/release-package_encoding-iso-8859-1.json'],
                     ['realdata/release-package_encoding-utf-8.json'])


def test_command_bad_encoding_iso_8859_1(capsys, monkeypatch, caplog):
    with caplog.at_level(logging.ERROR):
        assert_streaming_error(capsys, monkeypatch, main, ['echo'],
                               ['realdata/release-package_encoding-iso-8859-1.json'])

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'CRITICAL'
        # macOS and Linux have different error messages.
        assert re.search(r"^encoding error: (?:'utf-8' codec can't decode byte 0xd3 in position \d+: invalid "
                         r"continuation byte|lexical error: invalid bytes in UTF8 string.(\n[^\n]+){2}\n)\n"
                         r"Try `--encoding iso-8859-1`\?$", caplog.records[0].message)


def test_command_ascii(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['--ascii', 'echo'],
                     ['encoding_utf-8.json'], ['encoding_ascii.json'])


def test_command_utf_8(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['echo'],
                     ['encoding_ascii.json'], ['encoding_utf-8.json'])


def test_command_pretty(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['--pretty', 'echo'],
                     ['release_minimal.json'], ['release_minimal_pretty.json'])


def test_command_compact(capsys, monkeypatch):
    assert_streaming(capsys, monkeypatch, main, ['echo'],
                     ['release_minimal_pretty.json'], ['release_minimal.json'])


def test_command_multiline_input(capsys, monkeypatch):
    stdin = b'{\n  "releases": [\n    {\n      "ocid": "x",\n      "date": "2001-02-03T00:00:00Z"\n    }\n  ]\n}'

    actual = run_streaming(capsys, monkeypatch, main, ['echo'], stdin)

    assert actual.out == '{"releases":[{"ocid":"x","date":"2001-02-03T00:00:00Z"}]}\n'


def test_command_array_input(capsys, monkeypatch):
    actual = run_streaming(capsys, monkeypatch, main, ['echo'], ['release-packages.json'])

    assert actual.out == (
        '{"uri":"http://example.com/id/1","publisher":{"name":"Acme"},"publishedDate":"2001-02-03T04:05:07Z","releases":[{"ocid":"ocds-213czf-1","id":"1","date":"2001-02-03T04:05:06Z","tag":["planning"],"initiationType":"tender"}],"version":"1.1"}\n'  # noqa: E501
        '{"uri":"http://example.com/id/1","publisher":{"name":"Acme"},"publishedDate":"2001-02-03T04:05:07Z","releases":[{"ocid":"ocds-213czf-1","id":"1","date":"2001-02-03T04:05:06Z","tag":["planning"],"initiationType":"tender"}],"version":"1.1"}\n'  # noqa: E501
    )


def test_command_invalid_json(capsys, monkeypatch, caplog):
    with caplog.at_level(logging.ERROR):
        stdin = read('release-package_minimal.json', 'rb') + b'\n{\n'

        assert_streaming_error(capsys, monkeypatch, main, ['echo'], stdin,
                               expected=read('release-package_minimal.json'))

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'CRITICAL'
        assert caplog.records[0].message.startswith('JSON error: ')



import json
import sys
from copy import deepcopy

from ocdskit.commands.base import BaseCommand
from ocdskit.schema import add_validation_properties
from ocdskit.util import json_dump


class Command(BaseCommand):
    name = 'schema-strict'
    help = 'adds "minItems" and "uniqueItems" if an array, "minProperties" if an object and "minLength" if a ' \
           'string and "enum", "format" and "pattern" are not set'

    def add_arguments(self):
        self.add_argument('file', help='the schema file')
        self.add_argument('--no-unique-items', action='store_true',
                          help="""don't add "uniqueItems" properties to array fields""")
        self.add_argument('--check', action='store_true',
                          help='check the file for missing properties without modifying the file')

    def handle(self):
        with open(self.args.file) as f:
            schema = json.load(f)

        original = deepcopy(schema)
        add_validation_properties(schema, unique_items=not self.args.no_unique_items)

        if self.args.check:
            if schema != original:
                print(f'ERROR: {self.args.file} is missing validation properties', file=sys.stderr)
        else:
            with open(self.args.file, 'w') as f:
                json_dump(schema, f, indent=2)
                f.write('\n')
